//---------------------------------------------------------------------------

#ifndef DigitDetectorH
#define DigitDetectorH
//---------------------------------------------------------------------------
#include "math.h"
#include <complex>
#include <syncobjs.hpp>
using namespace std;
//---------------------------------------------------------------------------
struct TPhoneDigit
{
	char Digit ;
	bool IsDTMF ;
	int StartTimeTicks ;
	int StartPosition ;
	int EndPosition ;
};
//---------------------------------------------------------------------------
struct TPulse
{
	int AttackPoint ;
	int StartPoint ;
	int EndPoint ;
	int PulsePoint ;
	int EchoPoint ;
	int Delta ;
	float Level ;
	int SilenceLength ;
	bool IsNoise ;
	bool IsLeading ;
	char Magnitude ;
};
//---------------------------------------------------------------------------
typedef void __fastcall (__closure *TBufferQueryEvent)( TComponent *Sender, short * &Buffer, int &Length, bool &IsLastBuffer, int Position );
typedef void __fastcall (__closure *TNotifyDigitEvent)( TComponent *Sender, char Digit, unsigned TimeTicks, bool IsDTMF );
//---------------------------------------------------------------------------
class TDigitDetectorThread ;
//---------------------------------------------------------------------------

				/* * * * * * * * * * * * * * *
				 *   __        _   __   __   *
				 *  |    |    /_| |__  |__   *
				 *  |__  |__ /  |  __|  __|  *
				 *                           *
				 * * * * * * * * * * * * * * */

class PACKAGE TDigitDetector : public TComponent
{
private:
/*       ____________________________
 *				 _   _          _  ___  __
 *				|_) |_) | \  / /_|  |  |_
 *				|   | \ |  \/ /  |  |  |__
 *       ____________________________
 *
 */
//______Friends_________________________________________________________________________________________________________________________________________________
	friend class TDigitDetectorThread ;

//______Member Variables________________________________________________________________________________________________________________________________________
	int mDonePosition, mDigitCount ;
	int total_samples, total_ambience_samples, total_pulse_samples, mAmbienceCount ;
	double total_pulse_level, total_ambience_level, total_avg_level ;
	bool mIsDetecting, mIsIdentified, mPulseDetected, mHasFollower, mIdle, mNotifying, mDetectionEndNotified ;
	unsigned short int * mSampleDeltaBuffer;
	float *mSampleLevelBuffer ;
	float *mGraph;
	TPulse **mPulseList;
	TPhoneDigit **mDigitList ;
	int mStartPoint, mEndPoint, mPulseCount, mNextActivityPoint;
	complex<double> *mFFTBuffer ;
	double 	mAmbienceCoefficient, mPulseCoefficient, mCoefficient ;
	bool mAssumeIdentifiedOnToneDetect ;
	TDigitDetectorThread *mDetectorThread ;
	TMultiReadExclusiveWriteSynchronizer *mSynchronizer ;

//______Property Fields_________________________________________________________________________________________________________________________________________
	unsigned mCurrentTimeTicks ;
	int mCurrentPosition ;
	bool mIgnorePulseDigits ;
	bool mBlocking;
	TThreadPriority mPriority ;
	TNotifyDigitEvent FOnDigitDetected ;
	TBufferQueryEvent FDoSupplyBuffer ;
	TNotifyEvent FOnDetectionEnd;

//______Property Get/Set Methods________________________________________________________________________________________________________________________________
	void __fastcall SetBlocking( bool aValue ) ;
	bool __fastcall GetBlocking() ;
	void __fastcall SetCurrentPosition( int aValue ) ;
	int __fastcall  GetCurrentPosition() ;
	int __fastcall  GetCurrentTimeTicks() ;
	void __fastcall SetIgnorePulseDigits(bool aValue) ;
	bool __fastcall GetIgnorePulseDigits() ;
	void __fastcall SetPriority( TThreadPriority aValue ) ;
	TThreadPriority __fastcall GetPriority() ;
	int __fastcall GetDonePosition();
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


protected:
/*        _______________________________________
 *				 _   _   _  ___  __   __  ___  __   __
 *				|_) |_) | |  |  |_   |     |  |_   |  \
 *				|   | \ |_|  |  |__  |__   |  |__  |__/
 *        _______________________________________
 *
 */
//______Method Specifications___________________________________________________________________________________________________________________________________
	void __fastcall fft( complex<double> *x, int nu ) ;

	int __fastcall DetectDigits() ;

	bool __fastcall IsGranulityValid( short int *aBuffer, int aFirstExtremum, int aNextExtremum, int aDistance ) ;

	bool __fastcall IsVibration( short int *aBuffer, int aFirstExtremum, int aNextExtremum, int aDistance ) ;

	bool __fastcall IsPulse( short int *aBuffer, unsigned short int *aDeltaBuffer, int &aDelta, int &aStartPoint, int &aEndPoint, int &aImpulsePoint
		, int aSilenceLength, bool aIsTerminator, bool aIsFirst ) ;

	bool __fastcall ScanPulses( short int *aSampleBuffer, unsigned short int *aDeltaBuffer, float *aLevelBuffer, int &aStartPoint, int &aEndPoint
		, int aMinorDelta, int aMajorDelta, int aTimeOut, TPulse **aPulseList, int &aPulseCount) ;

	void __fastcall NotifyDigitDetected( char aDigit, unsigned aTimeTicks, bool aIsDTMF ) ;
	void __fastcall NotifyDetectionEnd() ;
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


public:
/*        _____________________
 *				 _       _         __
 *				|_) | | |_) |   | |
 *				|   |_| |_) |__ | |__
 *        _____________________
 *
 */
//______Constructor And Destructor______________________________________________________________________________________________________________________________
	__fastcall ~TDigitDetector();
	__fastcall TDigitDetector(TComponent* Owner);

//______Method Specifications___________________________________________________________________________________________________________________________________
	void __fastcall ResetSampleParams() ;

	void __fastcall GetSampleParams( unsigned short int *aDeltaBuffer, float *aLevelBuffer, short int *aSampleBuffer, int aLength ) ;

	bool __fastcall FindSoundSignal( int &aStartPoint, int &aEndPoint, int &aImpulsePoint, int &aNextActivityStartPoint
		, float *aLevelBuffer, unsigned short int *aDeltaBuffer, float aCorrectionCoefficient
		, float aMinCutLevel, float aMaxCutLevel, float aAttackMinLevelSlope, float aReleaseMinLevelSlope, bool aSplit
		, float aMinDeltaLevelRatio, float aMinLevelWidthRatio, float aMinHeightWidthRatio
		, int aMinWidth, float aMinDelta ) ;

	bool __fastcall DetectTone( int &aStartPoint, int &aEndPoint, char &aTone, short int * aBuffer, complex<double> * aFFTBuffer ) ;

	bool __fastcall DetectPulseDigit( int &aStartPoint, int &aEndPoint, char &aDigit, bool &aHasFollower, int &aAvgPulseWidth, int &aAvgDist, int &aPulseCount
		, short int *aSampleBuffer, unsigned short int *aDeltaBuffer, float *aLevelBuffer, TPulse **aPulseList, float *aGraph ) ;

	bool __fastcall BeginDetect(int aStartPosition, unsigned aStartTimeTicks, bool aHasFollower, bool aIdentify, bool aAssumeIdentifiedOnToneDetect );

	void __fastcall EndDetect();
	bool __fastcall ContinueDetect();

//______Property Definitions____________________________________________________________________________________________________________________________________
	__property int CurrentPosition  = { read=GetCurrentPosition, write=SetCurrentPosition };
	__property unsigned CurrentTimeTicks  = { read=GetCurrentTimeTicks };
	__property int DonePosition  = { read=GetDonePosition };
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


__published:
/*        ____________________________________
 *				 _       _         __        __  __
 *				|_) | | |_) |   | |__  |__| |_  |  \
 *				|   |_| |_) |__ |  __| |  | |__ |__/
 *        ____________________________________
 *
 */
//______Property Definitions____________________________________________________________________________________________________________________________________
	__property bool Blocking  = { read=GetBlocking, write=SetBlocking, default=true };
	__property bool IgnorePulseDigits  = { read=GetIgnorePulseDigits, write=SetIgnorePulseDigits };
	__property TThreadPriority Priority  = { read=GetPriority, write=SetPriority, default=tpLowest };

//______Event Definitions_______________________________________________________________________________________________________________________________________
	__property TNotifyDigitEvent OnDigitDetected  = { read=FOnDigitDetected, write=FOnDigitDetected };
	__property TBufferQueryEvent DoSupplyBuffer  = { read=FDoSupplyBuffer, write=FDoSupplyBuffer };
	__property TNotifyEvent OnDetectionEnd  = { read=FOnDetectionEnd, write=FOnDetectionEnd };
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
};
//---------------------------------------------------------------------------

				/* * * * * * * * * * * * * * *
				 *   __        _   __   __   *
				 *  |    |    /_| |__  |__   *
				 *  |__  |__ /  |  __|  __|  *
				 *                           *
				 * * * * * * * * * * * * * * */

class TDigitDetectorThread : public TThread
{
protected:
	TDigitDetector *mDigitDetector ;
	char mDigit ;
	unsigned mTimeTicks ;
	bool mIsDTMF ;
	void __fastcall NotifyDigitDetected() ;
	void __fastcall Execute() ;
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


public:
	__fastcall TDigitDetectorThread(bool CreateSuspended, TDigitDetector *aDigitDetector) ;
	__fastcall ~TDigitDetectorThread() ;
	void __fastcall DigitDetected(char aDigit, unsigned aTimeTicks, bool aIsDTMF) ;
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
};
//---------------------------------------------------------------------------
#endif
