//---------------------------------------------------------------------------
#ifndef WaveDeviceH
#define WaveDeviceH
//---------------------------------------------------------------------------
#ifndef MAX_QUEUE_ITEMS
#define MAX_QUEUE_ITEMS 10000
#endif
//---------------------------------------------------------------------------
#include <SysUtils.hpp>
#include <Controls.hpp>
#include <Classes.hpp>
#include <Forms.hpp>
//--------------------------------------------------------------------------------------------------------------------------------
#include "mmsystem.h"
//--------------------------------------------------------------------------------------------------------------------------------
enum TNotificationMethod {nmNone, nmEvent, nmCallbackFunction, nmThreadMessage, nmWindowMessage} ;
//--------------------------------------------------------------------------------------------------------------------------------
typedef void __fastcall (__closure *TBufferEvent)(TComponent *Sender, void *Buffer, int Length ) ;
typedef void __fastcall (__closure *TWaveInBufferEvent)(TComponent *Sender, void *Buffer, int BufferSize, int Length, int Position, int Tag) ;
typedef void __fastcall (__closure *TWaveOutBufferEvent)(TComponent *Sender, void *Buffer, int Length, int Position, int Tag) ;
typedef void __fastcall (__closure *TPopQueueEvent)(TComponent *Sender, bool &Accept, void *Buffer, int BufferSize, int Length, int Position, int Tag) ;
typedef void __fastcall (__closure *TSystemErrorEvent)(TComponent *Sender, char *ErrorText, int ErrorCode ) ;
//--------------------------------------------------------------------------------------------------------------------------------
class PACKAGE TWaveDevice : public TComponent
{
private:
/*				 _   _           _   ___  __
 *				|_) |_) | \  /  /_|   |  |_
 *				|   | \ |  \/  /  |   |  |__
 */
//______Member Variables________________________________________________________________________________________________________________________________________
	WAVEHDR **mWaveInHeaderQueue ;
	DWORD mWaveInOpenFlags ;
	WAVEFORMATEX  mWaveInFormat ;
	HWAVEIN mWaveInHandle ;
	HANDLE mWaveInEvent ;
	TMultiReadExclusiveWriteSynchronizer *mWaveInQueueSynchronizer ;		// to Synchronize modifications in WaveIn Buffer Queue.
	TMultiReadExclusiveWriteSynchronizer *mWaveInDeviceSynchronizer ;	// to Synchronize WaveIn process.
	TMultiReadExclusiveWriteSynchronizer *mWaveInCallbackSynchronizer ;
	int mWaveInCallbackOffset ;	// From mWaveInQueuePopIndex
	int mWaveInCallbackPosition ;
	int mWaveInCallbackInitialPosition ;
	bool mWaveInCallbackAbsolutePosition ;
	int mWaveInFirstPreparedIndex ;    // will be synchronized by Device synchronizer
	int mWaveInLastStopPosition ;	//**
	int mWaveInPreparedBuffersCount ;  // will be synchronized by Device synchronizer
	bool mWaveInStopped ;
	bool mWaveInUnprepareOnPop ;
	bool mWaveInUserCallBack ;
	int mWaveInDeviceCount ;
	WAVEINCAPS mWaveInCaps ;
	int mWaveInQueueAbsolutePosition ;
	int mWaveInQueueBaseIndex ;     //**
	int mWaveInQueueBasePosition ;  //**
	int mWaveInQueueBufferCount ;
	int mWaveInQueueFirstFreeIndex ;
	int mWaveInQueueInitialPosition ;
	int mWaveInQueuePopIndex ;		// The buffer will be popped from here
	int mWaveInQueuePopPosition ;
	int mWaveInQueuePushIndex ;		// The buffer will be pushed here
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	WAVEHDR **mWaveOutHeaderQueue ;
	DWORD mWaveOutOpenFlags ;
	WAVEFORMATEX  mWaveOutFormat ;
	HWAVEOUT mWaveOutHandle ;
	HANDLE mWaveOutEvent ;
	TMultiReadExclusiveWriteSynchronizer *mWaveOutQueueSynchronizer ;		// to Synchronize modifications in WaveOut Buffer Queue.
	TMultiReadExclusiveWriteSynchronizer *mWaveOutDeviceSynchronizer ;	// to Synchronize WaveOut process.
	TMultiReadExclusiveWriteSynchronizer *mWaveOutCallbackSynchronizer ;
	int mWaveOutCallbackOffset ;	// From mWaveOutQueuePopIndex
	int mWaveOutCallbackPosition ;
	int mWaveOutCallbackInitialPosition ;
	bool mWaveOutCallbackAbsolutePosition ;
	int mWaveOutFirstPreparedIndex ;    // will be synchronized by Device synchronizer
	int mWaveOutPreparedBuffersCount ;  // will be synchronized by Device synchronizer
	bool mWaveOutStopped ;
	bool mWaveOutUnprepareOnPop ;
	bool mWaveOutUserCallBack ;
	int mWaveOutDeviceCount ;
	WAVEOUTCAPS mWaveOutCaps ;
	int mWaveOutQueueAbsolutePosition ;
	int mWaveOutQueueBufferCount ;
	int mWaveOutQueueFirstFreeIndex ;
	int mWaveOutQueueInitialPosition ;
	int mWaveOutQueuePopIndex ;		// The buffer will be popped from here
	int mWaveOutQueuePopPosition ;
	int mWaveOutQueuePushIndex ;		// The buffer will be pushed here

//______Property Fields_________________________________________________________________________________________________________________________________________
	bool mWaveInAbsolutePosition ;
	bool mWaveInActive ;
	int mWaveInAvgBytesPerSec ;
	int mWaveInBitsPerSample ;
	int mWaveInBlockAlign ;
	LONG mWaveInCallBack ;
	int mWaveInChannels ;
	int mWaveInDeviceID ;
	bool mWaveInSampleAccurate ;
	TNotificationMethod mWaveInNotificationMethod ;
	int mWaveInSamplesPerSec ;
	TWaveInBufferEvent FOnWaveInBufferDone ;
	TSystemErrorEvent FOnWaveInError ;
	TPopQueueEvent FOnWaveInPop ;
	TBufferEvent FOnWaveInEnumerate ;
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	bool mWaveOutAbsolutePosition ;
	bool mWaveOutActive ;
	int mWaveOutAvgBytesPerSec ;
	int mWaveOutBitsPerSample ;
	int mWaveOutBlockAlign ;
	LONG mWaveOutCallBack ;
	int mWaveOutChannels ;
	int mWaveOutDeviceID ;
	TNotificationMethod mWaveOutNotificationMethod ;
	int mWaveOutSamplesPerSec ;
	TWaveOutBufferEvent FOnWaveOutBufferDone ;
	TSystemErrorEvent FOnWaveOutError ;
	TPopQueueEvent FOnWaveOutPop ;

//______Property Get/Set Methods________________________________________________________________________________________________________________________________
	void __fastcall SetWaveInActive(bool aActive) ;
	bool __fastcall GetWaveInActive() ;
	void __fastcall SetWaveInAbsolutePosition(bool aValue) ;
	bool __fastcall GetWaveInAbsolutePosition() ;
	void __fastcall SetWaveInAvgBytesPerSec(int aAvgBytesPerSec) ;
	int __fastcall GetWaveInAvgBytesPerSec() ;
	void __fastcall SetWaveInBitsPerSample(int aBitsPerSample) ;
	int __fastcall GetWaveInBitsPerSample() ;
	void __fastcall SetWaveInBlockAlign(int aBlockAlign) ;
	int __fastcall GetWaveInBlockAlign() ;
	void __fastcall SetWaveInCallBack(LONG aWaveCallBack) ;
	LONG __fastcall GetWaveInCallBack() ;
	void __fastcall SetWaveInChannels(int aChannels) ;
	int __fastcall GetWaveInChannels() ;
	void __fastcall SetWaveInDeviceID(int aDeviceID) ;
	int __fastcall GetWaveInDeviceID() ;
	AnsiString __fastcall GetWaveInDeviceName() ;
	void __fastcall SetWaveInSampleAccurate(bool value) ;
	bool __fastcall GetWaveInSampleAccurate() ;
	void __fastcall SetWaveInNotificationMethod(TNotificationMethod aNotificationMethod) ;
	TNotificationMethod __fastcall GetWaveInNotificationMethod() ;
	int __fastcall GetWaveInPosition() ;
	void __fastcall SetWaveInSamplesPerSec(int aSamplesPerSec) ;
	int __fastcall GetWaveInSamplesPerSec() ;
	void __fastcall SetWaveInStopped(bool aStopped) ;
	bool __fastcall GetWaveInStopped() ;
	void __fastcall SetWaveInUnprepareOnPop(bool aValue) ;
	bool __fastcall GetWaveInUnprepareOnPop() ;
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	void __fastcall SetWaveOutActive(bool aActive) ;
	bool __fastcall GetWaveOutActive() ;
	void __fastcall SetWaveOutAbsolutePosition(bool aValue) ;
	bool __fastcall GetWaveOutAbsolutePosition() ;
	void __fastcall SetWaveOutAvgBytesPerSec(int aAvgBytesPerSec) ;
	int __fastcall GetWaveOutAvgBytesPerSec() ;
	void __fastcall SetWaveOutBitsPerSample(int aBitsPerSample) ;
	int __fastcall GetWaveOutBitsPerSample() ;
	void __fastcall SetWaveOutBlockAlign(int aBlockAlign) ;
	int __fastcall GetWaveOutBlockAlign() ;
	void __fastcall SetWaveOutCallBack(LONG aWaveCallBack) ;
	LONG __fastcall GetWaveOutCallBack() ;
	void __fastcall SetWaveOutChannels(int aChannels) ;
	int __fastcall GetWaveOutChannels() ;
	void __fastcall SetWaveOutDeviceID(int aDeviceID) ;
	int __fastcall GetWaveOutDeviceID() ;
	bool __fastcall GetWaveOutSampleAccurate() ;
	AnsiString __fastcall GetWaveOutDeviceName() ;
	void __fastcall SetWaveOutNotificationMethod(TNotificationMethod aNotificationMethod) ;
	TNotificationMethod __fastcall GetWaveOutNotificationMethod() ;
	int __fastcall GetWaveOutPosition() ;
	void __fastcall SetWaveOutSamplesPerSec(int aSamplesPerSec) ;
	int __fastcall GetWaveOutSamplesPerSec() ;
	void __fastcall SetWaveOutStopped(bool aStopped) ;
	bool __fastcall GetWaveOutStopped() ;
	void __fastcall SetWaveOutUnprepareOnPop(bool aValue) ;
	bool __fastcall GetWaveOutUnprepareOnPop() ;

//______Method Specifications___________________________________________________________________________________________________________________________________
	bool __fastcall WaveInUnprepareBuffers() ;
	bool __fastcall WaveInBuffersUnprepared() ;
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	bool __fastcall WaveOutUnprepareBuffers() ;
	bool __fastcall WaveOutBuffersUnprepared() ;

//______Friends_________________________________________________________________________________________________________________________________________________
	friend void CALLBACK WaveInPrc(HWAVEIN hwi, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2) ;
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	friend void CALLBACK WaveOutPrc(HWAVEOUT hwo, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2) ;


protected:
/*				 _   _   _  ___  __   __  ___  __   __
 *				|_) |_) | |  |  |_   |     |  |_   |  \
 *				|   | \ |_|  |  |__  |__   |  |__  |__/
 */
//______Property Fields_________________________________________________________________________________________________________________________________________
	bool mHalfDuplex ;
	AnsiString mDummyString ;
	bool mDummyBool ;

//______Property Get/Set Methods________________________________________________________________________________________________________________________________
	void __fastcall SetHalfDuplex(bool aValue) ;
	bool __fastcall GetHalfDuplex() ;

//______Method Specifications___________________________________________________________________________________________________________________________________
	virtual bool __fastcall WaveInError(MMRESULT aResult) ;
	virtual bool __fastcall WaveInDone(void *Buffer, int BufferSize, int Length, int Tag) ;
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	virtual bool __fastcall WaveOutError(MMRESULT aResult) ;
	virtual bool __fastcall WaveOutDone(void *Buffer, int Length, int Tag) ;


public:
/*				 _       _         __
 *				|_) | | |_) |   | |
 *				|   |_| |_) |__ | |__
 */
//______Method Specifications___________________________________________________________________________________________________________________________________
	bool __fastcall WaveInPush(void *aWaveBuffer, int aBufferLength, int aTag) ;
	bool __fastcall WaveInPop(void* &aWaveBuffer, int &aBufferLength, int &aBytesRecorded, int &aPosition, int &aTag) ; //  - - - - -
	bool __fastcall WaveInClear() ;
	bool __fastcall WaveInReset() ;
	bool __fastcall WaveInStart() ;
	bool __fastcall WaveInStop() ;
	bool __fastcall WaveInWaitForNotify(int aTimeOutMilliseconds) ;
	bool __fastcall WaveInWriteTo(void *aBuffer, int &aStartPosition, int &aLength, bool aAbsolutePosition, bool aShiftIfTrimmed ) ;
	bool __fastcall WaveInEnumerate(int &aStartPosition, int &aLength, bool aAbsolutePosition) ;
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	bool __fastcall WaveOutPush(void *aWaveBuffer, int aBufferLength, int aTag) ;
	bool __fastcall WaveOutPop(void* &aWaveBuffer, int &aBufferLength, int &aBytesPlayed, int &aPosition, int &aTag) ;
	bool __fastcall WaveOutClear() ;
	bool __fastcall WaveOutReset() ;
	bool __fastcall WaveOutStart() ;
	bool __fastcall WaveOutStop() ;
	bool __fastcall WaveOutWaitForNotify(int aTimeOutMilliseconds) ;

//______Constructor And Destructor______________________________________________________________________________________________________________________________
	__fastcall TWaveDevice(TComponent* Owner) ;
	__fastcall ~TWaveDevice() ;

//______Property Definitions____________________________________________________________________________________________________________________________________
	__property LONG WaveInCallBack  = { read=GetWaveInCallBack, write=SetWaveInCallBack } ;
	__property int WaveInPosition  = { read=GetWaveInPosition } ;
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	__property LONG WaveOutCallBack  = { read=GetWaveOutCallBack, write=SetWaveOutCallBack } ;
	__property int WaveOutPosition  = { read=GetWaveOutPosition } ;


__published:
/*				 _       _         __        __  __
 *				|_) | | |_) |   | |__  |__| |_  |  \
 *				|   |_| |_) |__ |  __| |  | |__ |__/
 */
//______Property Definitions____________________________________________________________________________________________________________________________________
	__property bool HalfDuplex  = { read=GetHalfDuplex, write=SetHalfDuplex } ;
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	__property bool WaveInActive  = { read=GetWaveInActive, write=SetWaveInActive } ;
	__property bool WaveInAbsolutePosition = { read=GetWaveInAbsolutePosition, write=SetWaveInAbsolutePosition } ;
	__property int WaveInAvgBytesPerSec  = { read=GetWaveInAvgBytesPerSec, write=SetWaveInAvgBytesPerSec } ;
	__property int WaveInBitsPerSample  = { read=GetWaveInBitsPerSample, write=SetWaveInBitsPerSample } ;
	__property int WaveInBlockAlign  = { read=GetWaveInBlockAlign, write=SetWaveInBlockAlign } ;
	__property int WaveInChannels  = { read=GetWaveInChannels, write=SetWaveInChannels } ;
	__property int WaveInDeviceID  = { read=GetWaveInDeviceID, write=SetWaveInDeviceID } ;
	__property AnsiString WaveInDeviceName  = { read=GetWaveInDeviceName, write=mDummyString } ;
	__property bool WaveInSampleAccurate = { read=GetWaveInSampleAccurate, write=SetWaveInSampleAccurate } ;
	__property int WaveInSamplesPerSec  = { read=GetWaveInSamplesPerSec, write=SetWaveInSamplesPerSec } ;
	__property bool WaveInStopped = { read=GetWaveInStopped, write=SetWaveInStopped } ;
	__property TNotificationMethod WaveInNotificationMethod  = { read=GetWaveInNotificationMethod, write=SetWaveInNotificationMethod } ;
	__property bool WaveInUnprepareOnPop  = { read=GetWaveInUnprepareOnPop, write=SetWaveInUnprepareOnPop } ;
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	__property bool WaveOutActive  = { read=GetWaveOutActive, write=SetWaveOutActive } ;
	__property bool WaveOutAbsolutePosition = { read=GetWaveOutAbsolutePosition, write=SetWaveOutAbsolutePosition } ;
	__property int WaveOutAvgBytesPerSec  = { read=GetWaveOutAvgBytesPerSec, write=SetWaveOutAvgBytesPerSec } ;
	__property int WaveOutBitsPerSample  = { read=GetWaveOutBitsPerSample, write=SetWaveOutBitsPerSample } ;
	__property int WaveOutBlockAlign  = { read=GetWaveOutBlockAlign, write=SetWaveOutBlockAlign } ;
	__property int WaveOutChannels  = { read=GetWaveOutChannels, write=SetWaveOutChannels } ;
	__property int WaveOutDeviceID  = { read=GetWaveOutDeviceID, write=SetWaveOutDeviceID } ;
	__property AnsiString WaveOutDeviceName  = { read=GetWaveOutDeviceName, write=mDummyString } ;
	__property bool WaveOutSampleAccurate = { read=GetWaveOutSampleAccurate, write=mDummyBool } ;
	__property int WaveOutSamplesPerSec  = { read=GetWaveOutSamplesPerSec, write=SetWaveOutSamplesPerSec } ;
	__property bool WaveOutStopped = { read=GetWaveOutStopped, write=SetWaveOutStopped } ;
	__property TNotificationMethod WaveOutNotificationMethod  = { read=GetWaveOutNotificationMethod, write=SetWaveOutNotificationMethod } ;
	__property bool WaveOutUnprepareOnPop  = { read=GetWaveOutUnprepareOnPop, write=SetWaveOutUnprepareOnPop } ;

//______Event Definitions_______________________________________________________________________________________________________________________________________
	__property TWaveInBufferEvent OnWaveInBufferDone  = { read=FOnWaveInBufferDone, write=FOnWaveInBufferDone } ;
	__property TSystemErrorEvent OnWaveInError  = { read=FOnWaveInError, write=FOnWaveInError } ;
	__property TPopQueueEvent OnWaveInPop = { read=FOnWaveInPop, write = FOnWaveInPop } ;
	__property TBufferEvent OnWaveInEnumerate  = { read=FOnWaveInEnumerate, write=FOnWaveInEnumerate } ;
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	__property TWaveOutBufferEvent OnWaveOutBufferDone  = { read=FOnWaveOutBufferDone, write=FOnWaveOutBufferDone } ;
	__property TSystemErrorEvent OnWaveOutError  = { read=FOnWaveOutError, write=FOnWaveOutError } ;
	__property TPopQueueEvent OnWaveOutPop = { read = FOnWaveOutPop, write = FOnWaveOutPop } ;
} ;
//--------------------------------------------------------------------------------------------------------------------------------
#endif
